/**************************************************************************
*
*	Controller.h
*
*	Controller API class, which makes it possible to control robots!
*
*	Author: Saxon Druce
*
*	Copyright (c) 2004 Block Software Pty Ltd
*
**************************************************************************/

#ifndef _CONTROLLER_H
#define _CONTROLLER_H

/*************************************************************************/
// Include files
/*************************************************************************/

#include <vector>

#include "CommsClient.h"
#include "Vec.h"

/*************************************************************************/
// Defines
/*************************************************************************/

/*************************************************************************/
// Public types
/*************************************************************************/

// Controller_t class - the class used to control a robot
class Controller_t : private CommsClient_t
{
public:

	// SensedRobot_t class - holds details of a robot which has been sensed
	struct SensedRobot_t
	{
	public:

		// the robot's ID
		int m_RobotID;

		// the robot's position
		Vec_t m_Position;

		// the robot's orientation
		double m_Orientation;
	};

	// SensedMunition_t class - holds details of a munition which has been sensed
	struct SensedMunition_t
	{
	public:

		// the munition's ID
		int m_MunitionID;

		// the munition's position
		Vec_t m_Position;
	};

	// Constructor
	Controller_t();

	// Destructor
	~Controller_t();

	// Connect() - Connects to the specified port on the specified host
	// (name or IP address). Throws an exception on failure to connect.
	// void Connect(
	//	const std::string& HostName,
	//	int Port);
	using CommsClient_t::Connect; // pull into public use from the base class

	// GetPlayingFieldSize() - Gets the size (width and height) of the 
	// playing field, in metres. The origin (0, 0) is in the middle of 
	// the playing field, so the field extends from -1/2 to +1/2 of the 
	// width and height in the x and y directions respectively.
	void GetPlayingFieldSize(double& Width, double& Height);

	// CreateRobot() - Creates a robot, returning its ID used for future
	// calls into the API.
	int CreateRobot();

	// RemoveRobot() - Removes the robot of the specified ID from the 
	// simulation. It must be a robot which you have created. If you
	// disconnect while a robot you have created exists, it will be
	// automatically removed.
	void RemoveRobot(int RobotID);

	// SetName() - Sets the robot's name.
	void SetName(int RobotID, const std::string& Name);

	// SetOwner() - Sets the robot's owner.
	void SetOwner(int RobotID, const std::string& Owner);

	// Taunt() - Sends a message to the server which will be sent to all 
	// viewers.
	void Taunt(const std::string& Message);

	// GetMoney() - Returns the amount of money which the robot has
	// remaining for purchasing equipment.
	int GetMoney(int RobotID);

	// PurchaseEngine() - Purchases the specified engine for the robot.
	void PurchaseEngine(int RobotID, const std::string& Name);

	// PurchaseSensor() - Purchases the specified sensor for the robot.
	void PurchaseSensor(int RobotID, const std::string& Name);

	// PurchaseBattery() - Purchases the specified battery for the robot.
	void PurchaseBattery(int RobotID, const std::string& Name);

	// PurchaseTurret() - Purchases the specified turret for the robot.
	void PurchaseTurret(int RobotID, const std::string& Name);

	// PurchaseWeapon() - Purchases the specified weapon for the robot.
	void PurchaseWeapon(int RobotID, const std::string& Name);

	// PurchaseSolarPanel() - Purchases the specified solar panel for 
	// the robot.
	void PurchaseSolarPanel(int RobotID, const std::string& Name);

	// EnterSimulation() - Enters the robot into active simulation, by
	// placing it at a random location and orientation on the playing
	// field.
	void EnterSimulation(int RobotID);

	// GetHealth() - Returns the current health of the robot, a value from
	// 0 (dead) to 100 (full).
	double GetHealth(int RobotID);

	// GetPosition() - Returns the current position of the robot.
	Vec_t GetPosition(int RobotID);

	// GetOrientation() - Returns the current orientation of the robot,
	// ie the angle with the x-axis in radians, positive clockwise.
	double GetOrientation(int RobotID);

	// SetThrottle() - Sets the robot's throttle value, from -100 
	// (backwards at 100% of engine's top speed) to 100 (forwards at
	// 100% of engine's top speed).
	void SetThrottle(int RobotID, double Throttle);

	// SetDesiredOrientation() - Sets the robot's desired orientation,
	// ie the angle with the x-axis in radians, positive clockwise.
	// Once the desired orientation is set, the robot will turn as
	// fast as its engine allows from its current orientation to the
	// desired orientation.
	void SetDesiredOrientation(int RobotID, double DesiredOrientation);

	// GetSensorOrientation() - Returns the current orientation of the
	// sensor, in radians, where 0 means pointing in the same direction
	// as the robot, positive is clockwise (to the right).
	double GetSensorOrientation(int RobotID);

	// SetDesiredSensorOrientation() - Sets the robot's desired sensor
	// orientation, ie the angle relative to the robot's orientation,
	// positive clockwise. Once the desired orientation is set, the 
	// sensor will turn as fast as possible from its current orientation 
	// to the desired orientation.
	void SetDesiredSensorOrientation(
		int RobotID, double DesiredSensorOrientation);

	// GetSensedRobots() - Returns a list of all robots which the sensor
	// of the specified robot can currently see.
	std::vector<SensedRobot_t> GetSensedRobots(int RobotID);

	// GetSensedMunitions() - Returns a list of all munitions which the 
	// sensor of the specified robot can currently see.
	std::vector<SensedMunition_t> GetSensedMunitions(int RobotID);

	// GetEnergy() - Returns the robot's current energy level.
	double GetEnergy(int RobotID);

	// GetTurretOrientation() - Returns the current orientation of the
	// turret, in radians, where 0 means pointing in the same direction
	// as the robot, positive is clockwise (to the right).
	double GetTurretOrientation(int RobotID);

	// SetDesiredTurretOrientation() - Sets the robot's desired turret
	// orientation, ie the angle relative to the robot's orientation,
	// positive clockwise. Once the desired orientation is set, the 
	// turret will turn as fast as possible from its current orientation 
	// to the desired orientation.
	void SetDesiredTurretOrientation(
		int RobotID, double DesiredTurretOrientation);

	// GetWeaponCharge() - Returns the percentage that the weapon is
	// charged, from 0 to 100.
	double GetWeaponCharge(int RobotID);

	// FireWeapon() - Fires the robot's weapon.
	void FireWeapon(int RobotID);
};

/*************************************************************************/

#endif // _CONTROLLER_H
