/**************************************************************************
*
*	Main.cpp
*
*	Main source file for simple robot sample programme
*
*	Author: Saxon Druce
*
*	Copyright (c) 2004 Block Software Pty Ltd
*
**************************************************************************/

/*************************************************************************/
// Include files
/*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <time.h>

#include "Controller.h"

/*************************************************************************/
// Defines
/*************************************************************************/

/*************************************************************************/
// Local types
/*************************************************************************/

/*************************************************************************/
// Local functions
/*************************************************************************/

// RandomBetween() - Helper function to return a random value between the 
// two bounds.
static double RandomBetween(double LowerBound, double UpperBound);

/*************************************************************************/
// Functions
/*************************************************************************/

/**************************************************************************
*
*	function:	main()
*
*	desc:		Programme entry point
*
*	notes:		
*
**************************************************************************/

void main(int NumArgs, char** Args)
{
	try
	{
		printf("Simple Robot Sample\n\n");

		// seed random number generator
		srand((unsigned int)time(NULL));

		// get the server name from the arguments, or use localhost if
		// not provided
		std::string ServerName;
		if (NumArgs >= 2)
			ServerName = Args[1];
		else
			ServerName = "localhost";

		// create a controller
		Controller_t Controller;

		// connect to the server
		Controller.Connect(ServerName, 9386);

		// get the playing field size
		double PlayingFieldWidth, PlayingFieldHeight;
		Controller.GetPlayingFieldSize(
			PlayingFieldWidth, PlayingFieldHeight);
		printf("Playing field size is %.1fm x %.1fm\n", 
			PlayingFieldWidth, PlayingFieldHeight);

		// create a robot
		int RobotID = Controller.CreateRobot();
		printf("Robot ID is %d\n", RobotID);

		// set name and owner
		Controller.SetName(RobotID, "Simple Robot");
		Controller.SetOwner(RobotID, "Saxon");

		// get initial amount of money
		printf("Robot has $%d to spend on equipment\n", 
			Controller.GetMoney(RobotID));

		// buy equipment
		Controller.PurchaseEngine(RobotID, "Engine.2.50");
		Controller.PurchaseSensor(RobotID, "Sensor.10.60.40");
		Controller.PurchaseBattery(RobotID, "Battery.50");
		Controller.PurchaseTurret(RobotID, "Turret.0.0");
		Controller.PurchaseWeapon(RobotID, "Weapon.2.4.2.3");
		Controller.PurchaseSolarPanel(RobotID, "SolarPanel.1");

		// get amount of money left
		printf("Robot has $%d remaining after purchasing equipment\n", 
			Controller.GetMoney(RobotID));

		// enter the simulation
		Controller.EnterSimulation(RobotID);

		// set up a value to hold the position which the robot is
		// trying to get to, and initialise it with the robot's
		// current position
		Vec_t TargetPosition = Controller.GetPosition(RobotID);

		// now loop until a key is pressed
		printf("\nHit a key to stop\n\n");
		while (!kbhit())
		{
			// get the robot's current position
			Vec_t CurrentPosition = Controller.GetPosition(RobotID);

			// sense other robots
			std::vector<Controller_t::SensedRobot_t> Robots = 
				Controller.GetSensedRobots(RobotID);

			if (Robots.empty())
			{
				// we couldn't see any robots, so simply wander randomly
				// around the playing field, in the hope that we'll
				// eventually find another robot

				// work out the vector from the current position to the
				// target position
				Vec_t TargetVector = TargetPosition - CurrentPosition;

				if (TargetVector.Length() < 1.0)
				{
					// we're within 1 metre of the target, so assume we've
					// reached it - therefore pick a new random target
					// position, which is at least 2 metres from the edge
					// of the playing field
					TargetPosition.x = RandomBetween(
						-(PlayingFieldWidth / 2.0 - 2.0), 
						(PlayingFieldWidth / 2.0 - 2.0));
					TargetPosition.y = RandomBetween(
						-(PlayingFieldHeight / 2.0 - 2.0), 
						(PlayingFieldHeight / 2.0 - 2.0));

					printf(
						"New target position is %.1f, %.1f\n",
						TargetPosition.x,
						TargetPosition.y);
					
					// recalculate the target vector since it is used below
					TargetVector = TargetPosition - CurrentPosition;
				}

				// work out the desired orientation to point towards the
				// target
				double DesiredOrientation = TargetVector.GetAngle();

				// set it
				Controller.SetDesiredOrientation(
					RobotID,
					DesiredOrientation);

				// get the robot's current orientation
				double CurrentOrientation = Controller.GetOrientation(RobotID);

				// work out the difference between the current orientation and
				// the desired orientation
				double OrientationDifference = 
					DesiredOrientation - CurrentOrientation;
				
				// wrap to -pi to pi range
				if (OrientationDifference < -M_PI)
					OrientationDifference += 2 * M_PI;
				else if (OrientationDifference >= M_PI)
					OrientationDifference -= 2 * M_PI;

				if (fabs(OrientationDifference) < 5 * M_PI / 180)
				{
					// we're within 5 degrees of the desired orientation, so
					// set the throttle to full forwards
					Controller.SetThrottle(RobotID, 100.0);
				}
				else
				{
					// set the throttle to stay still while we turn
					Controller.SetThrottle(RobotID, 0.0);
				}

				// set sensor orientation to 90 degrees further clockwise
				// than it currently is, which should result in a constant
				// clockwise spin as fast as possible (assuming we aren't
				// able to spin by 90 degrees between iterations of this loop)
				Controller.SetDesiredSensorOrientation(
					RobotID, Controller.GetSensorOrientation(RobotID) + M_PI / 2);
			}
			else
			{
				// we can see at least one robot, so attack the first one

				// work out the vector from the current position to the
				// enemy robot position
				Vec_t EnemyVector = Robots[0].m_Position - CurrentPosition;

				// work out the desired orientation to point towards the
				// enemy robot
				double DesiredOrientation = EnemyVector.GetAngle();

				// set it
				Controller.SetDesiredOrientation(
					RobotID,
					DesiredOrientation);

				// get the robot's current orientation
				double CurrentOrientation = Controller.GetOrientation(RobotID);

				// work out the difference between the current orientation and
				// the desired orientation
				double OrientationDifference = 
					DesiredOrientation - CurrentOrientation;
				
				// wrap to -pi to pi range
				if (OrientationDifference < -M_PI)
					OrientationDifference += 2 * M_PI;
				else if (OrientationDifference >= M_PI)
					OrientationDifference -= 2 * M_PI;

				if (fabs(OrientationDifference) < 5 * M_PI / 180)
				{
					// we're within 5 degrees of the desired orientation

					// work out how far we are from the enemy
					double EnemyDistance = EnemyVector.Length();

					if (EnemyDistance < 5)
					{
						// within 5 metres of the enemy, so move
						// backwards to keep away from it (we'll
						// be too easy to hit if we get too close)
						Controller.SetThrottle(RobotID, -100.0);
					}
					else if (EnemyDistance < 6)
					{
						// between 5 and 6 metres from the enemy,
						// so stay still
						Controller.SetThrottle(RobotID, 0.0);
					}
					else
					{
						// more than 6 metres from the enemy, so
						// move towards it
						Controller.SetThrottle(RobotID, 100.0);
					}

					if (EnemyDistance < 6)
					{
						// we're within a reasonable range of the enemy,
						// so fire the weapon if it's fully charged
						if (Controller.GetWeaponCharge(RobotID) >= 100)
							Controller.FireWeapon(RobotID);
					}
				}
				else
				{
					// set the throttle to stay still while we turn
					Controller.SetThrottle(RobotID, 0.0);
				}

				// set the sensor angle required to keep it pointing 
				// at the enemy, to make sure it doesn't get away
				Controller.SetDesiredSensorOrientation(
					RobotID, DesiredOrientation - CurrentOrientation);
			}
		}
		// read the pressed key
		getch();

		printf("\nDone\n\n");
	}
	catch (std::exception& e)
	{
		printf("\nAn exception occurred: %s\n\n", e.what());
	}
}

/**************************************************************************
*
*	function:	RandomBetween()
*
*	desc:		Helper function to return a random value between the two 
*				bounds.
*
*	notes:		
*
**************************************************************************/

static double RandomBetween(double LowerBound, double UpperBound)
{
	return LowerBound + rand() * (UpperBound - LowerBound) / RAND_MAX;
}

/*************************************************************************/
