/**************************************************************************
*
*	SimpleRobot.cs
*
*	Main source file for simple robot sample programme (in C#)
*
*	Author: Saxon Druce
*
*	Copyright (c) 2004 Block Software Pty Ltd
*
**************************************************************************/

using System;
using System.Collections;
using ControllerNet;

namespace SimpleRobotNet
{
	/// <summary>
	/// The main class.
	/// </summary>
	class SimpleRobot
	{
		/// <summary>
		/// random number generator
		/// </summary>
		static Random m_random = new Random();

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			try
			{
				// get the server name from the arguments, or use localhost if
				// not provided
				string serverName;
				if (args.Length >= 1)
					serverName = args[0];
				else
					serverName = "localhost";

				// create a controller
				Controller controller = new Controller();

				// connect
				controller.Connect(serverName, 9386);

				// get the playing field size
				double PlayingFieldWidth, PlayingFieldHeight;
				controller.GetPlayingFieldSize(
					out PlayingFieldWidth, out PlayingFieldHeight);
				Console.WriteLine("Playing field size is {0}m x {1}m", 
					PlayingFieldWidth, PlayingFieldHeight);

				// create a robot
				int robotID = controller.CreateRobot();
				Console.WriteLine("Robot ID is {0}", robotID);

				// set name and owner
				controller.SetName(robotID, "Simple Robot C#");
				controller.SetOwner(robotID, "Saxon");

				// get initial amount of money
				Console.WriteLine("Robot has ${0} to spend on equipment", 
					controller.GetMoney(robotID));

				// buy equipment
				controller.PurchaseEngine(robotID, "Engine.2.50");
				controller.PurchaseSensor(robotID, "Sensor.10.60.40");
				controller.PurchaseBattery(robotID, "Battery.50");
				controller.PurchaseTurret(robotID, "Turret.0.0");
				controller.PurchaseWeapon(robotID, "Weapon.2.4.2.3");
				controller.PurchaseSolarPanel(robotID, "SolarPanel.1");

				// get amount of money left
				Console.WriteLine("Robot has ${0} remaining after purchasing equipment", 
					controller.GetMoney(robotID));

				// enter the simulation
				controller.EnterSimulation(robotID);

				// set up a value to hold the position which the robot is
				// trying to get to, and initialise it with the robot's
				// current position
				Vec targetPosition = controller.GetPosition(robotID);

				// loop forever, since there doesn't seem to be a way
				// to loop until a key is pressed
				Console.WriteLine("\nUse ctrl+C to stop\n");
				for (;;)
				{
					// get the robot's current position
					Vec currentPosition = controller.GetPosition(robotID);

					// sense other robots
					ArrayList robots = controller.GetSensedRobots(robotID);

					if (robots.Count == 0)
					{
						// we couldn't see any robots, so simply wander randomly
						// around the playing field, in the hope that we'll
						// eventually find another robot

						// work out the vector from the current position to the
						// target position
						Vec targetVector = targetPosition - currentPosition;

						if (targetVector.Length() < 1.0)
						{
							// we're within 1 metre of the target, so assume we've
							// reached it - therefore pick a new random target
							// position, which is at least 2 metres from the edge
							// of the playing field
							targetPosition.x = RandomBetween(
								-(PlayingFieldWidth / 2.0 - 2.0), 
								(PlayingFieldWidth / 2.0 - 2.0));
							targetPosition.y = RandomBetween(
								-(PlayingFieldHeight / 2.0 - 2.0), 
								(PlayingFieldHeight / 2.0 - 2.0));

							Console.WriteLine(
								"New target position is {0:f1}, {0:f1}",
								targetPosition.x,
								targetPosition.y);
					
							// recalculate the target vector since it is used below
							targetVector = targetPosition - currentPosition;
						}

						// work out the desired orientation to point towards the
						// target
						double desiredOrientation = targetVector.GetAngle();

						// set it
						controller.SetDesiredOrientation(
							robotID,
							desiredOrientation);

						// get the robot's current orientation
						double currentOrientation = controller.GetOrientation(robotID);

						// work out the difference between the current orientation and
						// the desired orientation
						double orientationDifference = 
							desiredOrientation - currentOrientation;
				
						// wrap to -pi to pi range
						if (orientationDifference < -Math.PI)
							orientationDifference += 2 * Math.PI;
						else if (orientationDifference >= Math.PI)
							orientationDifference -= 2 * Math.PI;

						if (Math.Abs(orientationDifference) < 5 * Math.PI / 180)
						{
							// we're within 5 degrees of the desired orientation, so
							// set the throttle to full forwards
							controller.SetThrottle(robotID, 100.0);
						}
						else
						{
							// set the throttle to stay still while we turn
							controller.SetThrottle(robotID, 0.0);
						}

						// set sensor orientation to 90 degrees further clockwise
						// than it currently is, which should result in a constant
						// clockwise spin as fast as possible (assuming we aren't
						// able to spin by 90 degrees between iterations of this loop)
						controller.SetDesiredSensorOrientation(
							robotID, 
							controller.GetSensorOrientation(robotID) + Math.PI / 2);
					}
					else
					{
						// we can see at least one robot, so attack the first one

						// work out the vector from the current position to the
						// enemy robot position
						Vec enemyVector = 
							(robots[0] as SensedRobot).m_Position - currentPosition;

						// work out the desired orientation to point towards the
						// enemy robot
						double desiredOrientation = enemyVector.GetAngle();

						// set it
						controller.SetDesiredOrientation(
							robotID,
							desiredOrientation);

						// get the robot's current orientation
						double currentOrientation = controller.GetOrientation(robotID);

						// work out the difference between the current orientation and
						// the desired orientation
						double orientationDifference = 
							desiredOrientation - currentOrientation;
				
						// wrap to -pi to pi range
						if (orientationDifference < -Math.PI)
							orientationDifference += 2 * Math.PI;
						else if (orientationDifference >= Math.PI)
							orientationDifference -= 2 * Math.PI;

						if (Math.Abs(orientationDifference) < 5 * Math.PI / 180)
						{
							// we're within 5 degrees of the desired orientation

							// work out how far we are from the enemy
							double enemyDistance = enemyVector.Length();

							if (enemyDistance < 5)
							{
								// within 5 metres of the enemy, so move
								// backwards to keep away from it (we'll
								// be too easy to hit if we get too close)
								controller.SetThrottle(robotID, -100.0);
							}
							else if (enemyDistance < 6)
							{
								// between 5 and 6 metres from the enemy,
								// so stay still
								controller.SetThrottle(robotID, 0.0);
							}
							else
							{
								// more than 6 metres from the enemy, so
								// move towards it
								controller.SetThrottle(robotID, 100.0);
							}

							if (enemyDistance < 6)
							{
								// we're within a reasonable range of the enemy,
								// so fire the weapon if it's fully charged
								if (controller.GetWeaponCharge(robotID) >= 100)
									controller.FireWeapon(robotID);
							}
						}
						else
						{
							// set the throttle to stay still while we turn
							controller.SetThrottle(robotID, 0.0);
						}

						// set the sensor angle required to keep it pointing 
						// at the enemy, to make sure it doesn't get away
						controller.SetDesiredSensorOrientation(
							robotID, desiredOrientation - currentOrientation);
					}
				}
			}
			catch (Exception e)
			{
				Console.WriteLine(
					"An exception occurred: {0}",
					e.Message);
			}
		}

		/// <summary>
		/// Helper function to return a random value between the two bounds.
		/// </summary>
		static double RandomBetween(double lowerBound, double upperBound)
		{
			return lowerBound + m_random.NextDouble() * (upperBound - lowerBound);
		}
	}
}
